/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.aop;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.jboss.aop.InstanceAdvisor;
import org.jboss.aop.Advised;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.joinpoint.Invocation;
import org.jboss.aop.joinpoint.MethodInvocation;
import org.jboss.aop.proxy.ClassProxy;
import org.jboss.aop.proxy.ClassProxyFactory;
import org.jboss.aop.util.MethodHashing;
import org.jboss.util.NestedRuntimeException;
import org.jboss.cache.Fqn;
import org.jboss.cache.CacheException;
import org.jboss.cache.GlobalTransaction;

import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;
import java.io.Serializable;

/**
 * AopCache delegation to handle internal cache sotre, that is, the portion that is not part of user's data.
 *
 * @author Ben Wang
 */
public class InternalDelegate
{
   static Log log=LogFactory.getLog(InternalDelegate.class.getName());
   public static final String CLASS_INTERNAL = "__jboss:internal:class__";
   public static final Fqn JBOSS_INTERNAL = new Fqn("__JBossInternal__");
   // We will try to break up the AopInstance inside the cache store.
   public static final String AOPINSTANCE_REF_COUNT = "__AOPINSTANCE_COUNT__";
   public static final String AOPINSTANCE_REF_FQN = "__AOPINSTANCE_FQN__";
   // This is the key to store the pojo instance. But this is not replicated instead local only.
   public static final String AOPINSTANCE_POJO = "__AOPINSTANCE_POJO__";

   protected TreeCacheAop cache_;

   InternalDelegate(TreeCacheAop cache)
   {
      cache_ = cache;
   }

//   Object getObjectInstance(AOPInstance aopInstance) {
//      return aopInstance.get();
//   }

   void resetRefCount(Fqn fqn) throws CacheException
   {
      cache_.put(fqn, AOPINSTANCE_REF_COUNT, new Integer(1));
   }

   /**
    * Increment reference count for the pojo. Note that this is not thread safe or atomic.
    */
   int incrementRefCount(Fqn fqn) throws CacheException
   {
      /**
       * We used to bundle ref count and fqn with aop instance. But now split them up to avoid
       * the class loading problem.
       */
      // Needs to be atomic.
      Integer count = (Integer)cache_.peek(fqn, AOPINSTANCE_REF_COUNT);
      if( count == null )
         throw new RuntimeException("InternalDelegate.incrementRefCount(): REF_COUNT for reference counting is null at: " +fqn);

      int counter = count.intValue() +1;
      cache_.put(fqn, AOPINSTANCE_REF_COUNT, new Integer(counter));
      return counter;
   }

   /**
    * decrement reference count for the pojo. Note that this is not thread safe or atomic.
    */
   int decrementRefCount(Fqn fqn) throws CacheException
   {
      // Needs to be atomic.
      Integer count = (Integer)cache_.peek(fqn, AOPINSTANCE_REF_COUNT);
      if( count == null )
         throw new RuntimeException("InternalDelegate.decrementRefCount(): REF_COUNT for reference counting is null at: " +fqn);

      int counter = count.intValue();
      if(counter == 0)
         throw new RuntimeException("InternalDelegate.decrementRefCount(): REF_COUNT for reference counting is 0 at: " +fqn);

      counter--;
      cache_.put(fqn, AOPINSTANCE_REF_COUNT, new Integer(counter));
      return counter;
   }

   String getRefFqn(Fqn fqn) throws CacheException {
      return (String)cache_.peek(fqn, AOPINSTANCE_REF_FQN);
   }

   void setRefFqn(Fqn fqn, String internalFqn) throws CacheException
   {
      cache_.put(fqn, AOPINSTANCE_REF_FQN, internalFqn);
   }

   void removeRefFqn(Fqn fqn) throws CacheException
   {
      cache_.remove(fqn, AOPINSTANCE_REF_FQN);
   }

   Object getObjectInstance(Fqn fqn) throws CacheException
   {
      return getPOJO(fqn);
   }

   Object getPOJO(Fqn fqn) throws CacheException
   {
      return cache_.peek(fqn, AOPINSTANCE_POJO);
   }

   void setPOJO(Fqn fqn, Object obj) throws CacheException
   {
      // Should do this in TreeCache.
      GlobalTransaction tx = cache_.getCurrentTransaction();
      cache_._put(tx, fqn, AOPINSTANCE_POJO, obj, true);
   }

   /**
    * We store the class name in string.
    */
   void putAopClazz(Fqn fqn, Class clazz) throws CacheException {
      cache_.marshalledPut(fqn, CLASS_INTERNAL, clazz);
   }

   Class peekAopClazz(Fqn fqn) throws CacheException {
      return (Class)cache_.marshalledGet(fqn, CLASS_INTERNAL);
   }

   public static boolean isInternalNode(Fqn fqn)
   {
     if(fqn.isChildOf(JBOSS_INTERNAL)) return true;
     return false;
   }

   // TODO Need to mangle the name to obtain uniqueness?
   Fqn createInternalNode(Fqn storedFqn)
   {
      Fqn fqn = new Fqn(JBOSS_INTERNAL, storedFqn);
      return fqn;
   }

   public boolean isAopNode(Fqn fqn) throws CacheException
   {
      // We use REF_COUNT as a marker to denote aop node now.
      if( cache_.peek(fqn, AOPINSTANCE_REF_COUNT) != null )
         return true;
      else
         return false;
   }

}
