/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.aop;

import org.jboss.aop.Advised;
import org.jboss.aop.ClassInstanceAdvisor;
import org.jboss.aop.InstanceAdvisor;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.proxy.ClassProxy;
import org.jboss.cache.*;
import org.jboss.cache.eviction.AopEvictionPolicy;
import org.jboss.cache.transaction.BatchModeTransactionManager;
import org.jboss.util.NestedRuntimeException;
import org.jgroups.JChannel;

import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.TransactionManager;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.*;

/**
 * Implementation of the cache using dyanmic aop interceptors. The basic idea is that the caller only
 * uses the {@link #putObject(String,Object)}, {@link #getObject(String)} and
 * {@link #removeObject(String)} methods.<br>
 * When putting an object into the cache, the cache essentially takes care of how the object
 * will be replicated. It 'breaks' the object apart to map it onto the tree, and adds an
 * interceptor recursively for each object reachable from the root object. Those interceptor
 * 'know' to which part of the tree they are mapped. Whenever there is state change
 * ("set*" interceptors), the tree is modified using the superclass. All "get*" operations
 * to the object are intercepted and redirected to the tree.<br>
 * Bottom line is that a user simply puts <em>any</em> object into the cache, and the object
 * will be replicated to all caches in the cluster. The only requirement on cachable objects
 * is that they access all state through getters and setters.
 *
 * @author Harald Gliebe
 * @author Ben Wang
 */
public class TreeCacheAop extends MarshalledTreeCache implements TreeCacheAopMBean, TreeCacheAopIfc
{
   // Class -> CachedType
   // use WeakHashMap to allow class reloading
   protected Map cachedTypes = new WeakHashMap();
   public static final String DUMMY = "dummy";
   // Use batch mode tm to simulate the automicity.
   TransactionManager localTm_ = BatchModeTransactionManager.getInstance();
   protected TreeCacheAopDelegate delegate_;

   public TreeCacheAop(String cluster_name,
                       String props,
                       long state_fetch_timeout)
         throws Exception
   {
      super(cluster_name, props, state_fetch_timeout);
      init();
   }

   public TreeCacheAop() throws Exception
   {
      init();
   }

   public TreeCacheAop(JChannel channel) throws Exception
   {
      super(channel);
      init();
   }

   protected void init()
   {
      setMarshalling(false);
      delegate_ = new TreeCacheAopDelegate(this);
   }

   /**
    * Over-ride to make sure we are using an eviction policy specific to aop.
    *
    */
   public void setEvictionPolicyClass(String eviction_policy_class) {
      this.eviction_policy_class=eviction_policy_class;
      if(eviction_policy_class == null || eviction_policy_class.length() ==0)
         return;

      try {
         Object obj = getClass().getClassLoader().loadClass(eviction_policy_class).newInstance();
         if( ! (obj instanceof AopEvictionPolicy))
            throw new RuntimeException("TreeCacheAop.setEvictionPolicyClass(). Eviction policy provider:" +
                  eviction_policy_class +" is not an instance of AopEvictionPolicy.");
         eviction_policy_provider =(TreeCacheListener) obj;
         this.addTreeCacheListener(eviction_policy_provider );
      }
      catch(RuntimeException ex) {
         log.error("setEvictionPolicyClass(): failed creating instance of  " + eviction_policy_class, ex);
         throw ex;
      }
      catch(Throwable t) {
         log.error("setEvictionPolicyClass(): failed creating instance of  " + eviction_policy_class, t);
      }
   }

   /** Override to provide aop specific eviction.
    *
    * <p>
    * Called by eviction policy provider. Note that eviction is done only in local mode,
    * that is, it doesn't replicate the node removal. This is will cause the replcation nodes
    * not synchronizing, but it is ok since user is supposed to add the node again when get is
    * null. After that, the contents will be in sync.
    * @param fqn Will remove everythign assoicated with this fqn.
    * @throws CacheException
    */
   public void evict(Fqn fqn) throws CacheException {
      // We will remove all children nodes as well since we assume all children nodes are part
      // of this "object" node.
      if(delegate_.isAopNode(fqn)) {
         if(log.isDebugEnabled()) {
            log.debug("evict(): evicting whole aop node " +fqn);
         }
//         _remove(null, fqn, create_undo_ops, false);
         // TODO Why do we not want to remove the interceptor??? I think we should.
         // Because if we remove it, the caller has no idea that it has been taken off.
         // She will still think that caching is in effect. But this returns null value from cache.
         // What's is a good policy?
         boolean removeCacheInterceptor = false;
         _removeObject(fqn, removeCacheInterceptor);
      } else {
         super.evict(fqn);
      }
   }

   /**
    * Obtain a cache aop type for user to traverse the defined "primitive" types in aop.
    *
    * @param clazz The original pojo class
    * @return CachedType
    */
   public synchronized CachedType getCachedType(Class clazz)
   {
      CachedType type = (CachedType) cachedTypes.get(clazz);
      if (type == null) {
         type = new CachedType(clazz);
         cachedTypes.put(clazz, type);
      }
      return type;
   }

   /**
    * Retrieve the aop-enabled object from the cache.
    *
    * @param fqn String name that associates with this node.
    * @return Current content value. Null if does not exist.
    * @throws CacheException
    */
   public Object getObject(String fqn) throws CacheException
   {
      return getObject(Fqn.fromString(fqn));
   }

   /**
    * Retrieve the aop-enabled object from the cache. Return null if object does not exist in the cache.
    *
    * @param fqn Instance that associates with this node.
    * @return Current content value. Null if does not exist.
    * @throws CacheException
    */
   public Object getObject(Fqn fqn) throws CacheException
   {
      if( hasCurrentTransaction() )  // We have a transaction context going on now.
      {
         Object obj = _getObject(fqn);
         return obj;
      } else
      {
         // Start a new transaction, we need transaction so the operation is atomic.
         // getObject may not need it in the future.
         try
         {
            localTm_.begin();
            // TODO Need to find a btter way to lock the node.
            put(fqn, DUMMY, DUMMY);
            Object obj = _getObject(fqn);
            remove(fqn, DUMMY);
            return obj;
         }
         catch (Exception e)
         {
            e.printStackTrace();
            try
            {
               localTm_.setRollbackOnly();
            }
            catch (Exception exn)
            {
               exn.printStackTrace();
            }
            // We will need to alert Tomcat of this exception.
            throw new NestedRuntimeException("TreeCacheAop.getObject(): ", e);
         }
         finally
         {
            endTransaction();
         }

      }
   }

   /**
    * Insert an aop-enabled object into the cache.
    * It will also recursively put the any sub-object that is
    * declared as aop-capable (i.e., in <code>jboss-aop.xml</code>).
    * Note that <code>List</code>, <code>Map</code>, <code>Set</code>
    * attributes are aop-enabled, by default, as well.
    *
    * @param fqn The fqn string name to associate with the object in the cache.
    * @param obj aop-enabled object to be inerted into the cache. If null,
    *            it will nullify the fqn node.
    * @param obj Return the previous content under fqn.
    * @throws CacheException
    */
   public Object putObject(String fqn, Object obj) throws CacheException
   {
      return putObject(Fqn.fromString(fqn), obj);
   }

   /**
    * Insert an aop-enabled object into the cache.
    * It will also recursively put the any sub-object that is
    * declared as aop-capable (i.e., in <code>jboss-aop.xml</code>).
    * Note that <code>List</code>, <code>Map</code>, <code>Set</code>
    * attributes are aop-enabled, by default, as well.
    *
    * @param fqn The fqn instance to associate with the object in the cache.
    * @param obj aop-enabled object to be inerted into the cache. If null,
    *            it will nullify the fqn node.
    * @param obj Return the previous content under fqn.
    * @throws CacheException
    */
   public Object putObject(Fqn fqn, Object obj) throws CacheException
   {
      if( hasCurrentTransaction() )  // We have a transaction context going on now.
      {
         return _putObject(fqn, obj);
      } else
      {
         // Start a new transaction, we need transaction so the operation is atomic.
         try
         {
            localTm_.begin();
            // TODO Need to find a btter way to lock the node.
            put(fqn, DUMMY, DUMMY);
            Object objOld = _putObject(fqn,obj);
            remove(fqn, DUMMY);
            return objOld;
         }
         catch (Exception e)
         {
            e.printStackTrace();
            try
            {
               localTm_.setRollbackOnly();
            }
            catch (Exception exn)
            {
               exn.printStackTrace();
            }
            throw new NestedRuntimeException("TreeCacheAop.putObject(): ", e);
         }
         finally
         {
            endTransaction();
         }
      }
   }


   /**
    * Remove aop-enabled object from the cache.
    *
    * @param fqn String name that associates with this node.
    * @return Value object from this node.
    * @throws CacheException
    */
   public Object removeObject(String fqn) throws CacheException
   {
      return removeObject(Fqn.fromString(fqn));
   }

   /**
    * Remove aop-enabled object from the cache.
    *
    * @param fqn Instance that associates with this node.
    * @return Original value object from this node.
    * @throws CacheException
    */
   public Object removeObject(Fqn fqn) throws CacheException
   {
      if( hasCurrentTransaction() )  // We have a transaction context going on now.
      {
         return _removeObject(fqn, true);
      } else
      {
         // Start a new transaction, we need transaction so the operation is atomic.
         try
         {
            localTm_.begin();
            // TODO Need to find a btter way to lock the node.
            put(fqn, DUMMY, DUMMY);
            Object obj = _removeObject(fqn, true);
            remove(fqn, DUMMY);
            return obj;
         }
         catch (Exception e)
         {
            e.printStackTrace();
            try
            {
               localTm_.setRollbackOnly();
            }
            catch (Exception exn)
            {
               exn.printStackTrace();
            }
            // We will need to alert Tomcat of this exception.
            throw new NestedRuntimeException("TreeCacheAop.removeObject(): ", e);
         }
         finally
         {
            endTransaction();
         }
      }
   }

   protected boolean hasCurrentTransaction()
   {
      try {
         if(getCurrentTransaction() != null || localTm_.getTransaction() != null)
         {
            // We have transaction context. Return null to signify don't do anything
            return true;
         }
      } catch (SystemException e) {
         e.printStackTrace();
      }
      return false;
   }

   protected void endTransaction()
   {
      try {
         if(localTm_.getTransaction().getStatus() != Status.STATUS_MARKED_ROLLBACK)
         {
            localTm_.commit();
         } else
         {
            localTm_.rollback();
         }
      } catch (Exception e) {
         e.printStackTrace();
         throw new NestedRuntimeException("TreeCacheAop.endTransaction(): ", e);
      }
   }

   protected Object _getObject(Fqn fqn) throws CacheException
   {
      return delegate_._getObject(fqn);
   }

   /**
    */
   protected Object _putObject(Fqn fqn, Object obj) throws CacheException
   {
      return delegate_._putObject(fqn, obj);
   }

   protected Object _removeObject(Fqn fqn, boolean removeCacheInterceptor) throws CacheException
   {
      return delegate_._removeObject(fqn, removeCacheInterceptor);
   }

}

