/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.aop;

import org.jboss.aop.Advised;
import org.jboss.aop.ClassInstanceAdvisor;
import org.jboss.aop.InstanceAdvisor;
import org.jboss.aop.advice.Interceptor;
import org.jboss.aop.proxy.ClassProxy;
import org.jboss.cache.*;
import org.jboss.cache.eviction.AopEvictionPolicy;
import org.jboss.cache.transaction.BatchModeTransactionManager;
import org.jboss.util.NestedRuntimeException;
import org.jgroups.JChannel;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import javax.transaction.Status;
import javax.transaction.SystemException;
import javax.transaction.TransactionManager;
import java.io.Serializable;
import java.lang.reflect.Field;
import java.util.*;

/**
 * @author Ben Wang
 */
public class TreeCacheAopDelegate
{
   protected TreeCacheAop cache_;
   protected final static Log log=LogFactory.getLog(TreeCacheAopDelegate.class);
   protected InternalDelegate internal_;

   public TreeCacheAopDelegate(TreeCacheAop cache)
   {
      cache_ = cache;
      internal_ = new InternalDelegate(cache);
   }

   protected Object _getObject(Fqn fqn) throws CacheException
   {
      Object pojo = internal_.getObjectInstance(fqn);
      if ( pojo != null) {
         // we already have an advised instance
         return pojo;
      }

      // the class attribute is implicitly stored as an immutable read-only attribute
      Class clazz = internal_.peekAopClazz(fqn);
      //  clazz and aopInstance can be not null if this node is the replicated brother node.
      if (clazz == null)
         return null;

      CachedType type = cache_.getCachedType(clazz);
      Object obj = null;
      if (Advised.class.isAssignableFrom(clazz)) {
         String refFqn = internal_.getRefFqn(fqn);
         if (refFqn == null) { // Create a new instance and also add in the CacheInterceptor for this fqn.
            try {
               obj = clazz.newInstance();
            }
            catch(Exception e) {
               throw new CacheException("failed creating instance of " + clazz.getName(), e);
            }
            // Insert interceptor at runtime
            InstanceAdvisor advisor = ((Advised) obj)._getInstanceAdvisor();
            advisor.appendInterceptor(new CacheInterceptor(cache_, fqn, type));
         } else {
            // this is recursive. Need to obtain the object from parent fqn
            // No need to add CacheInterceptor as a result. Everything is re-directed.
            // In addition, this op will not be recursive.
            if (log.isDebugEnabled()) {
               log.debug("getObject(): obtain value from reference fqn: " + refFqn);
            }
            obj = cache_.getObject(refFqn);
         }

      } else { // Must be Collection classes. We will use aop.ClassProxy instance instead.
         try {
            if(Map.class.isAssignableFrom(clazz)) {
               obj=CollectionInterceptorUtil.createProxy(clazz, new CachedMapInterceptor(cache_, fqn, clazz));
            } else if(List.class.isAssignableFrom(clazz)) {
               obj=CollectionInterceptorUtil.createProxy(clazz, new CachedListInterceptor(cache_, fqn, clazz));
            } else if(Set.class.isAssignableFrom(clazz)) {
               obj=CollectionInterceptorUtil.createProxy(clazz, new CachedSetInterceptor(cache_, fqn, clazz));
            } else {
               // Maybe it is just a serialized object.
               obj=cache_.peek(fqn, "value");
            }
         }
         catch(Exception e) {
            throw new CacheException("failure creating proxy", e);
         }
      }

      internal_.setPOJO(fqn, obj);
      return obj;
   }

   protected Object retrieveCurrentValue(Fqn fqn, Object obj) throws CacheException {
      Object oldValue = null;
      Object pojo = internal_.getObjectInstance(fqn);
      if (pojo != null) {
         if (pojo == obj) {
            // obj is already cached. We are done.
            return obj;
         }

         // Optimization that won't remove the underlying object if
         // they are the same object type.
         log.debug("putObject(): old class type: " + pojo);
         if (obj.getClass().isInstance(pojo)) {
            if (log.isDebugEnabled()) {
               log.debug("putObject(): obj is same class type as the old one");
            }
            oldValue = cache_.getObject(fqn);
         } else {
            oldValue = cache_.removeObject(fqn);
         }
      } else {
         // If obj is a primitive type or other un-delcared classes, or the first time around.
         Class claz1 = internal_.peekAopClazz(fqn);
         if (claz1 != null &&
               obj.getClass().getName().equals(claz1.getName())) {
            if (log.isDebugEnabled()) {
               log.debug("putObject(): obj is same primitive class type as the old one");
            }
         }
      }
      return oldValue;
   }

   /**
    *
    * @param fqn
    * @param obj
    * @return
    * @throws CacheException
    */
   protected Object _putObject(Fqn fqn, Object obj) throws CacheException
   {
      AopUtil.checkObjectType(obj);
      if (obj == null) {
         return cache_.removeObject(fqn);
      }
      // Skip some un-necessary update if obj is the same class as the old one
      Object oldValue = retrieveCurrentValue(fqn, obj);
      if(oldValue == obj) return obj;  // value already in cache. return right away.

      // Remember not to print obj here since it will trigger the CacheInterceptor.
      if(log.isDebugEnabled()) {
         log.debug("putObject(): fqn: " + fqn);
      }

      // store object in cache
      if (obj instanceof Advised) {
         CachedType type = cache_.getCachedType(obj.getClass());
         // add interceptor
         InstanceAdvisor advisor = ((Advised) obj)._getInstanceAdvisor();
         if(advisor == null)
            throw new RuntimeException("_putObject(): InstanceAdvisor is null for: " +obj);

         // Let's check for object graph, e.g., multiple and circular references first
         if(handleCircularReference(fqn, advisor, obj, type))
         {
            return oldValue;
         } else if (handleObjectGraph(fqn, advisor, type, obj)) { // found cross references
            return oldValue;
         }

         _regularPutObject(fqn, obj, advisor, type);

         /**
          * Handling collection classes here.
          * First check if obj has been aspectized? That is, if it is a ClassProxy or not.
          * If not, we will need to create a proxy first for the Collection classes
          */
      } else if (_collectionPutObject(fqn, obj)) {
        //
      } else if (obj instanceof Serializable) { // must be Serializable, including pritimive types
         if (log.isDebugEnabled()) {
            log.debug("putObject(): obj (" + obj.getClass() + ") is non-advisable but is Serializable. ");
         }

         internal_.putAopClazz(fqn, obj.getClass());

         internal_.setPOJO(fqn, obj);
         // TODO Can also use obj.getName()?
         cache_.put(fqn, "value", obj);
      }

      return oldValue;
   }

   protected void _regularPutObject(Fqn fqn, Object obj, InstanceAdvisor advisor, CachedType type) throws CacheException
   {
      // TODO ? workaround for deserialiased objects
      if (advisor == null) {
         advisor = new ClassInstanceAdvisor(obj);
         ((Advised) obj)._setInstanceAdvisor(advisor);
      }

      // Insert interceptor at runtime
      advisor.appendInterceptor(new CacheInterceptor(cache_, fqn, type));
      // Always initialize the ref count so we can mark this as an AopNode.
      internal_.resetRefCount(fqn);
      // Let's deflate the objet here. If the field is another "aspectized" object,
      // we will do it recursively.
      internal_.putAopClazz(fqn, type.getType());

      for (Iterator i = type.getFields().iterator(); i.hasNext();) {
         Field field = (Field) i.next();
         Object value = null;
         try {
            value=field.get(obj);
         }
         catch(IllegalAccessException e) {
            throw new CacheException("field access failed", e);
         }
         CachedType fieldType = cache_.getCachedType(field.getType());
         if (fieldType.isImmediate()) {
            cache_.put(fqn, field.getName(), value);
         } else {
            Fqn tmpFqn = new Fqn(fqn, field.getName());
            _putObject(tmpFqn, value);
            // If value (field member) is of Collection type, e.g., composite class
            // that contains Collection member, we will swap out the old reference
            // with the proxy one.
            // This can probably be optimized with check for instanceof proxy
            if( value instanceof Map || value instanceof List || value instanceof Set ) {
               Object newValue = cache_.getObject(tmpFqn);
               try {
                  field.set(obj, newValue);
               } catch (IllegalAccessException e) {
                  log.error("_putObject(): Can't swap out the Collection class of field " +field.getName() +
                        "Exception " +e);
                  throw new CacheException("_putObject(): Can't swap out the Collection class of field \" +field.getName(),"
                        +e);
               }
            }
         }
      }

      // Put AOPInstance. Note pojo is "transient".
      internal_.setPOJO(fqn, obj);
      // Need to make sure this is behind put such that obj.toString is done correctly.
      if (log.isDebugEnabled()) {
         log.debug("putObject(): inserting with fqn: " + fqn.toString());
      }
   }

   protected boolean _collectionPutObject(Fqn fqn, Object obj) throws CacheException {
      boolean isCollection = false;

      if (obj instanceof Map) {
         if (log.isDebugEnabled()) {
            log.debug("putObject(): aspectized obj is a Map type of size: " + ((Map) obj).size());
         }

         // Need to remove the existing ones first.
         cache_.removeObject(fqn);
         internal_.putAopClazz(fqn, obj.getClass());

         // Let's replace it with a proxy if necessary
         Object oldObj = obj;
         if( !(obj instanceof ClassProxy)) {
            Class clazz = obj.getClass();
            try {
               obj=CollectionInterceptorUtil.createProxy(clazz, new CachedMapInterceptor(cache_, fqn, clazz));
            } catch (Exception e) {
               throw new CacheException("failure creating proxy", e);
            }
         }

         Map map = (Map) oldObj;

         for (Iterator i = map.entrySet().iterator(); i.hasNext();) {
            Map.Entry entry = (Map.Entry) i.next();
            _putObject(new Fqn(fqn, entry.getKey()), entry.getValue());
         }

         internal_.setPOJO(fqn, obj);
         isCollection = true;
      } else if (obj instanceof List) {
         if (log.isDebugEnabled()) {
            log.debug("putObject(): aspectized obj is a List type of size: "
                  + ((List) obj).size());
         }

         List list = (List) obj;
         // Need to remove the existing ones first.
         cache_.removeObject(fqn);
         internal_.putAopClazz(fqn, obj.getClass());

         // Let's replace it with a proxy if necessary
         if( !(obj instanceof ClassProxy)) {
            Class clazz = obj.getClass();
            try {
               obj=CollectionInterceptorUtil.createProxy(clazz, new CachedListInterceptor(cache_, fqn, clazz));
            } catch (Exception e) {
               throw new CacheException("failure creating proxy", e);
            }
         }

         int idx = 0;
         for (Iterator i = list.iterator(); i.hasNext();) {
            _putObject(new Fqn(fqn, new Integer(idx++)), i.next());
         }
         //        } else if (CachedType.isImmediate(obj.getClass())) {

         internal_.setPOJO(fqn, obj);
         isCollection = true;
      } else if (obj instanceof Set) {
         if (log.isDebugEnabled()) {
            log.debug("putObject(): aspectized obj is a Set type of size: "
                  + ((Set) obj).size());
         }

         Set set = (Set) obj;
         // Need to remove the existing ones first.
         cache_.removeObject(fqn);
         internal_.putAopClazz(fqn, obj.getClass());

         // Let's replace it with a proxy if necessary
         if( !(obj instanceof ClassProxy)) {
            Class clazz = obj.getClass();
            try {
               obj=CollectionInterceptorUtil.createProxy(clazz, new CachedSetInterceptor(cache_, fqn, clazz));
            } catch (Exception e) {
               throw new CacheException("failure creating proxy", e);
            }
         }

         int idx = 0;
         for (Iterator i = set.iterator(); i.hasNext();) {
            _putObject(new Fqn(fqn, new Integer(idx++)), i.next());
         }
         //        } else if (CachedType.isImmediate(obj.getClass())) {

         internal_.setPOJO(fqn, obj);
         isCollection = true;
      }

      return isCollection;
   }

   protected Object _removeObject(Fqn fqn, boolean removeCacheInterceptor) throws CacheException
   {
      Class clazz = internal_.peekAopClazz(fqn);
      if (clazz == null)
      {
         if (log.isTraceEnabled()) {
            log.trace("removeObject(): clasz is null. fqn: " + fqn);
         }
         return null;
      }

      if (log.isDebugEnabled()) {
         log.debug("removeObject(): removing object from fqn: " + fqn);
      }

      Object result = cache_.getObject(fqn);
      if(result == null)
      {
         if(cache_.exists(fqn))
            cache_.remove(fqn);   // brute force.

         return null;
      }

//      AOPInstance aopInstance = internal_.peekAopInstance(fqn);
      if (Advised.class.isAssignableFrom(clazz)) {
         _regularRemoveObject(fqn, removeCacheInterceptor, result, clazz);
      } else if (_collectionRemoveObject(fqn, removeCacheInterceptor)) {
      }

      // TODO kind of brute force now.
      cache_.remove(fqn);
      // remove the interceptor as well.
      return result;
   }

   protected void _regularRemoveObject(Fqn fqn, boolean removeCacheInterceptor, Object result, Class clazz) throws CacheException
   {
      String refFqn = internal_.getRefFqn(fqn);
      InstanceAdvisor advisor = ((Advised) result)._getInstanceAdvisor();
      // check if this is a refernce
      if (refFqn != null) {
         if (log.isDebugEnabled()) {
            log.debug("removeObject(): removing object fqn: " + fqn + " but is actually from ref fqn: " + refFqn);
         }
         removeRefFqn(fqn, refFqn, removeCacheInterceptor);
      } else {
         CachedType type = cache_.getCachedType(clazz);
         for (Iterator i = type.getFields().iterator(); i.hasNext();) {
            Field field = (Field) i.next();
            CachedType fieldType = cache_.getCachedType(field.getType());
            if (!fieldType.isImmediate()) {
               Object obj = _removeObject(new Fqn(fqn, field.getName()), removeCacheInterceptor);
               if (obj == null) continue;
            }
         }
         // Determine if we want to keep the interceptor for later use.
         if(removeCacheInterceptor) {
            CacheInterceptor interceptor = (CacheInterceptor) AopUtil.findCacheInterceptor(advisor);
//               if (log_.isDebugEnabled()) {
//                  log_.debug("removeObject(): removing cache interceptor fqn: " + fqn);
//               }
            // Remember to remove the interceptor from in-memory object but make sure it belongs to me first.
            if (interceptor != null)
            {
               if (log.isDebugEnabled()) {
                  log.debug("removeObject(): removed cache interceptor fqn: " + fqn + " interceptor: "+interceptor);
               }
               advisor.removeInterceptor(interceptor.getName());
            }
         }
      }

   }

   protected boolean _collectionRemoveObject(Fqn fqn, boolean removeCacheInterceptor) throws CacheException
   {
      boolean isCollection = false;
      Class clazz = internal_.peekAopClazz(fqn);

      if (Map.class.isAssignableFrom(clazz)) {
         Map values = cache_.get(fqn).getChildren();
         if (values != null) {
            ArrayList list = new ArrayList(values.keySet());   // need to clone it first.
            for (int i=0; i < list.size(); i++) {
               Object key = list.get(i);
               _removeObject(new Fqn(fqn, key), removeCacheInterceptor);
            }
         }
         isCollection = true;
      } else if (Collection.class.isAssignableFrom(clazz)) {
         Map values = cache_.get(fqn).getChildren();
         int size = values == null ? 0 : values.size();
         for (int i = 0; i < size; i++) {
            _removeObject(new Fqn(fqn, new Integer(i)), removeCacheInterceptor);
         }
         isCollection = true;
      }
      return isCollection;
   }

   /**
    *  Check for circular reference. If there is one, we will remove any object from current fqn (if any),
    * and then put a reference internal fqn in the AOPInstance in the current fqn.
    */
   protected boolean handleCircularReference(Fqn fqn, InstanceAdvisor advisor, Object obj, CachedType type)
      throws CacheException
   {
      Fqn internalFqn;
      if ((internalFqn = checkCircularReference(fqn, advisor, obj)) != null) {
         // 1. Remove existing one, if there is any. 2. set the ref fqn to the internal one,
         // 3. put aopInstane in the cache fqn node such that getObject will be redirected to the internal node.
         cache_.removeObject(fqn);
         this.addRefFqn(fqn, internalFqn.toString());
         internal_.putAopClazz(fqn, type.getType());
         return true;
      }
      else {
         return false;
      }

   }

   protected Fqn checkCircularReference(Fqn fqn, InstanceAdvisor advisor, Object obj)
   {
//      if (advisor == null) return null; // this can be a case of non-advised POJO.

      Fqn originalFqn = null;
      // Step Check for cross references
      Interceptor interceptor = AopUtil.findCacheInterceptor(advisor);
      if (interceptor == null) {
         return null;
      }

      // ah, found something. So this will be multiple referenced.
      originalFqn = ((CacheInterceptor) interceptor).getFqn();

      // if it is a circular reference,
      // we will simply return the fqn associated that
      // We know it is circular reference because it is the fqn child references back to parent.
      if (fqn.isChildOf(originalFqn)) {
         if (log.isDebugEnabled()) {
            log.debug("checkCircularReference(): is child for circular ref fqn " + originalFqn);
         }
         return originalFqn;
      }
      return null;
   }

   /**
    * Handle cases where there is circular, e.g., parent refers to child and child to parent, or multiple references,
    * e.g., two objects reference the same
    * sub-object. In this case, we will remove the current sub-object and substitue with a ref. The
    * reference will point to an Jboss internal node where the sub-object is re-created and stored.
    * It will also have reference counting to keep track of garbage collection.
    *
    * @param fqn     Current fqn to store this sub-object
    * @param advisor The associated Advisor instance
    * @return The fqn that refers to the sub-object stored in JBoss internal node.
    */
   protected boolean handleObjectGraph(Fqn fqn, InstanceAdvisor advisor, CachedType type, Object obj) throws CacheException
   {
      Fqn internalFqn = getObjectGraphInternalFqn(fqn, advisor, type, obj);
      if(internalFqn == null) return false;

      // Will need to remove the current object under fqn first.
      cache_.removeObject(fqn);

      // This will increment the ref count, reset, and add ref fqn in the current fqn node.
      this.addRefFqn(fqn, internalFqn.toString());
      internal_.putAopClazz(fqn, type.getType());

      return true;
   }

   protected Fqn getObjectGraphInternalFqn(Fqn fqn, InstanceAdvisor advisor, CachedType type, Object obj) throws CacheException {
      if (advisor == null) return null; // this can be a case of non-advised POJO.

      Fqn originalFqn = null;
      Fqn internalFqn = null;
      // Step Check for cross references
      Interceptor interceptor = AopUtil.findCacheInterceptor(advisor);
      if (interceptor == null) {
         if (log.isDebugEnabled()) {
            log.debug("handleMultipleReference(): No multiple refernce found for fqn: " + fqn);
         }
         // Nothing found. Means this is a fresh object. No need to handle.
         return null;
      }

      // ah, found something. So this will be multiple referenced.
      originalFqn = ((CacheInterceptor) interceptor).getFqn();
       // If we are the same fqn, this is not multiple referenced!
      // This can be the case when we will reconstruct the node behind the scence.
      if( fqn.equals(originalFqn) ) return null;

      if (log.isDebugEnabled()) {
         log.debug("handleObjectGraph(): Found multiple refernce at original fqn: " + originalFqn);
      }

      // Check if the reference node locates under JBoss internal already.
      // Or if it is a circular reference.
      // If it is, we will simply return the fqn associated that
      if (internal_.isInternalNode(originalFqn)) {
         if (log.isDebugEnabled()) {
            log.debug("handleObjectGraph(): is child for fqn " + originalFqn);
         }

         // Remove existing one if any
         cache_.removeObject(fqn);
         return originalFqn;
      } else {
         // If not, we will need to create the sub-object under JBoss internal
         internalFqn = internal_.createInternalNode(originalFqn);
         // Then we will remove the original sub-object node (leaves only AopInstance)
         Object oldValue = cache_.removeObject(originalFqn);
         // Then put it under JBoss internal
         cache_.putObject(internalFqn, obj);

         // Need to call this after putObject so interceptor will be there.
         this.addRefFqn(originalFqn, internalFqn.toString());
         internal_.putAopClazz(originalFqn, type.getType());

         if (log.isDebugEnabled()) {
            log.debug("handleObjectGraph(): relocate the original fqn: " + originalFqn +
                  " to JBossInternal: " + internalFqn + " with obj: " + oldValue);
//                log_.debug("handleObjectGraph(): print cache nodes details: " +printDetails());
         }
         // Finally, we return with the refFqn.
         return internalFqn;
      }
   }

   /**
    * 1. increment reference counter
    * 2. set the current ref counter (@fqn) to 1. Used for aop marker.
    * 3. put in refFqn so we can get it.
    */
   void addRefFqn(Fqn fqn, String refFqn) throws CacheException {
      synchronized (this) {
         // increment the reference counting
         internal_.incrementRefCount(Fqn.fromString(refFqn));
         internal_.resetRefCount(fqn); // marker for aop node.
         // set the internal fqn in fqn so we can reference it.
         internal_.setRefFqn(fqn, refFqn);
      }
   }

   void removeRefFqn(Fqn fqn, String refFqn, boolean removeCacheInterceptor) throws CacheException {
      synchronized (this) {
         // take care of reference counting
         // Idea is to track reference counting only for object stored in JBossInternal.
         if (internal_.decrementRefCount(Fqn.fromString(refFqn)) == 0) {
            _removeObject(Fqn.fromString(refFqn), removeCacheInterceptor);
         } else {
            internal_.removeRefFqn(fqn);
         }
      }

      internal_.removeRefFqn(fqn);
   }

   boolean isAopNode(Fqn fqn)
   {
      try {
         return (internal_.isAopNode(fqn));
      } catch (Exception e) {
         log.warn("isAopNode(): cache get operation generated exception " +e);
         return false;
      }
   }
}
