/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.aop;

import org.jboss.cache.CacheException;
import org.jboss.cache.Fqn;

/**
 * Interface for TreeCacheAop. Will need to consolidate with TreeCache interface in the future.
 * </p>
  * Implementation of the cache using dyanmic aop interceptors. The basic idea is that the caller only
  * uses the {@link #putObject(String,Object)}, {@link #getObject(String)} and
  * {@link #removeObject(String)} methods.<br>
  * When putting an object into the cache, the cache essentially takes care of how the object
  * will be replicated. It 'breaks' the object apart to map it onto the tree, and adds an
  * interceptor recursively for each object reachable from the root object. Those interceptor
  * 'know' to which part of the tree they are mapped. Whenever there is state change
  * ("set*" interceptors), the tree is modified using the superclass. All "get*" operations
  * to the object are intercepted and redirected to the tree.<br>
  * Bottom line is that a user simply puts <em>any</em> object into the cache, and the object
  * will be replicated to all caches in the cluster. The only requirement on cachable objects
  * is that they access all state through getters and setters.
  *
  * @author Harald Gliebe
  * @author Ben Wang
 */
public interface TreeCacheAopIfc {
   /**
     * Retrieve the aop-enabled object from the cache. Return null if object does not exist in the cache.
     *
     * @param fqn Instance that associates with this node.
     * @return Current content value. Null if does not exist.
     * @throws CacheException
     */
   Object getObject(String fqn) throws CacheException;

   /**
     * Retrieve the aop-enabled object from the cache. Return null if object does not exist in the cache.
     *
     * @param fqn Instance that associates with this node.
     * @return Current content value. Null if does not exist.
     * @throws CacheException
     */
   Object getObject(Fqn fqn) throws CacheException;

   /**
    * Insert an aop-enabled object into the cache.
    * It will also recursively put the any sub-object that is
    * declared as aop-capable (i.e., in <code>jboss-aop.xml</code>).
    * Note that <code>List</code>, <code>Map</code>, <code>Set</code>
    * attributes are aop-enabled, by default, as well.
    *
    * @param fqn The fqn instance to associate with the object in the cache.
    * @param obj aop-enabled object to be inerted into the cache. If null,
    *            it will nullify the fqn node.
    * @param obj Return the previous content under fqn.
    * @throws CacheException
    */
   Object putObject(String fqn, Object obj) throws CacheException;

   /**
    * Insert an aop-enabled object into the cache.
    * It will also recursively put the any sub-object that is
    * declared as aop-capable (i.e., in <code>jboss-aop.xml</code>).
    * Note that <code>List</code>, <code>Map</code>, <code>Set</code>
    * attributes are aop-enabled, by default, as well.
    *
    * @param fqn The fqn instance to associate with the object in the cache.
    * @param obj aop-enabled object to be inerted into the cache. If null,
    *            it will nullify the fqn node.
    * @param obj Return the previous content under fqn.
    * @throws CacheException
    */
   Object putObject(Fqn fqn, Object obj) throws CacheException;

   /**
    * Remove aop-enabled object from the cache.
    *
    * @param fqn Instance that associates with this node.
    * @return Original value object from this node.
    * @throws CacheException
    */
   Object removeObject(String fqn) throws CacheException;

   /**
    * Remove aop-enabled object from the cache.
    *
    * @param fqn Instance that associates with this node.
    * @return Original value object from this node.
    * @throws CacheException
    */
   Object removeObject(Fqn fqn) throws CacheException;
}
