package org.jboss.cache.loader.tcp;

import org.jboss.cache.TreeCache;
import org.jboss.cache.PropertyConfigurator;
import org.jboss.cache.Fqn;
import org.jboss.cache.Node;
import org.jboss.cache.loader.DelegatingCacheLoader;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.net.ServerSocket;
import java.net.Socket;
import java.io.*;
import java.util.Map;
import java.util.Set;

/**
 * TCP-IP based CacheServer, configure TcpDelegatingCacheLoader with host and port of this server
 * @author Bela Ban
 * @version $Id: TcpCacheServer.java,v 1.4 2005/04/28 14:30:06 bela Exp $
 */
public class TcpCacheServer {
   ServerSocket srv_sock;
   InetAddress  bind_addr=null;
   int          port=2099;
   TreeCache    cache;
   String       config;
   static Log   log=LogFactory.getLog(TcpCacheServer.class);


   public TcpCacheServer(String bind_addr, int port, String config) throws UnknownHostException {
      if(bind_addr != null)
         this.bind_addr=InetAddress.getByName(bind_addr);
      this.port=port;
      this.config=config;
   }

   public void start() throws Exception {
      Socket client_sock;
      Connection conn;

      cache=new TreeCache();
      PropertyConfigurator cfg=new PropertyConfigurator();
      cfg.configure(cache, config);
      cache.createService();
      cache.startService();

      srv_sock=new ServerSocket(port, 10, bind_addr);
      System.out.println("TcpCacheServer listening on : " + srv_sock.getInetAddress() + ":" + srv_sock.getLocalPort());
      while(true) {
         client_sock=srv_sock.accept();
         conn=new Connection(client_sock, cache);
         conn.start();
      }
   }



   private static class Connection implements Runnable {
      Socket             sock=null;
      ObjectInputStream  input=null;
      ObjectOutputStream output=null;
      TreeCache          c;
      Thread             t=null;

      public Connection(Socket sock, TreeCache cache) throws IOException {
         this.sock=sock;
         output=new ObjectOutputStream(sock.getOutputStream());
         input=new ObjectInputStream(sock.getInputStream());
         c=cache;
      }


      public void start() {
         t=new Thread(this, "TcpCacheServer.Connection");
         t.setDaemon(true);
         t.start();
      }

      public void run() {
         int op;
         Fqn fqn;
         Object key, val, retval;
         Map map;
         Node n;
         boolean flag;
         byte[] state;

         while(true) {
            try {
               op=input.readInt();
               switch(op) {
                  case DelegatingCacheLoader.delegateGetChildrenNames:
                     fqn=(Fqn)input.readObject();
                     Set children=c.getChildrenNames(fqn);
                     output.writeObject(children);
                     output.flush();
                     break;
                  case DelegatingCacheLoader.delegateGetKey:
                     fqn=(Fqn)input.readObject();
                     key=input.readObject();
                     retval=c.get(fqn, key);
                     output.writeObject(retval);
                     output.flush();
                     break;
                  case DelegatingCacheLoader.delegateGet:
                     fqn=(Fqn)input.readObject();
                     n=c.get(fqn);
                     map=n != null? n.getData() : null;
                     output.writeObject(map);
                     output.flush();
                     break;
                  case DelegatingCacheLoader.delegateExists:
                     fqn=(Fqn)input.readObject();
                     flag=c.exists(fqn);
                     output.writeBoolean(flag);
                     output.flush();
                     break;
                  case DelegatingCacheLoader.delegatePutKeyVal:
                     fqn=(Fqn)input.readObject();
                     key=input.readObject();
                     val=input.readObject();
                     retval=c.put(fqn, key, val);
                     output.writeObject(retval);
                     output.flush();
                     break;
                  case DelegatingCacheLoader.delegatePut:
                     fqn=(Fqn)input.readObject();
                     map=(Map)input.readObject();
                     c.put(fqn, map);
                     break;
                  case DelegatingCacheLoader.delegateRemoveKey:
                     fqn=(Fqn)input.readObject();
                     key=input.readObject();
                     retval=c.remove(fqn, key);
                     output.writeObject(retval);
                     output.flush();
                     break;
                  case DelegatingCacheLoader.delegateRemove:
                     fqn=(Fqn)input.readObject();
                     c.remove(fqn);
                     break;
                  case DelegatingCacheLoader.delegateRemoveData:
                     fqn=(Fqn)input.readObject();
                     c.removeData(fqn);
                     break;
                  case DelegatingCacheLoader.delegateLoadEntireState:
                     state=c.getCacheLoader() != null? c.getCacheLoader().loadEntireState() : null;
                     output.writeObject(state);
                     output.flush();
                     break;
                  case DelegatingCacheLoader.delegateStoreEntireState:
                     state=(byte[])input.readObject();
                     if(c.getCacheLoader() != null)
                        c.getCacheLoader().storeEntireState(state);
                     break;
                  default:
                     log.error("Operation " + op + " unknown");
                     break;
               }
            }
            catch(Exception e) {
               log.warn("failed reading data, thread will terminate", e);
               try {if(output != null) output.close();} catch(Throwable th) {}
               try {if(input != null) input.close();} catch(Throwable th) {}
               try {if(sock != null) sock.close();} catch(Throwable th) {}
               break;
            }
         }
      }
   }


   public static void main(String[] args) throws Exception {
      String bind_addr=null;
      int port=2099;
      TcpCacheServer server;
      String config="cache-service.xml";

      for(int i=0; i < args.length; i++) {
         if(args[i].equals("-bind_addr")) {
            bind_addr=args[++i];
            continue;
         }
         if(args[i].equals("-port")) {
            port=Integer.parseInt(args[++i]);
            continue;
         }
         if(args[i].equals("-config")) {
            config=args[++i];
            continue;
         }
         help();
         return;
      }
      server=new TcpCacheServer(bind_addr, port, config);
      server.start();
   }



   private static void help() {
      System.out.println("TcpCacheServer [-bind_addr <address>] [-port <port>] [-help]");
   }
}
