/*
 * JBoss, the OpenSource J2EE webOS
 *
 * Distributable under LGPL license.
 * See terms of license at gnu.org.
 */
package org.jboss.cache.loader.tcp;

import org.jboss.cache.Fqn;
import org.jboss.cache.TreeCache;
import org.jboss.cache.loader.DelegatingCacheLoader;

import java.io.IOException;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.net.Socket;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

/**
 * DelegatingCacheLoader implementation which delegates to a remote (not in the same VM)
 * TreeCache using TCP/IP for communication. Example configuration for connecting to a TcpCacheServer
 * running at myHost:12345:<pre>
 * <attribute name="CacheLoaderClass">org.jboss.cache.loader.tcp.TcpDelegatingCacheLoader</attribute>
   <attribute name="CacheLoaderConfig">
       host=localhost
       port=2099
   </attribute>
 </pre>
 *
 * @author Bela Ban
 * @version $Id: TcpDelegatingCacheLoader.java,v 1.2 2005/04/28 13:44:27 bela Exp $
 */
public class TcpDelegatingCacheLoader extends DelegatingCacheLoader {
   private Socket     sock;
   private String     host;
   private int        port;
   ObjectInputStream  in;
   ObjectOutputStream out;


   /**
    * Default constructor.
    */
   public TcpDelegatingCacheLoader() {
      // Empty.
   }

   /**
    * Allows programmatic configuration.
    * 
    * @param cache The cache that the cacheloader will acting on behalf of.
    * @param host The host on which to look up the remote object.
    * @param port The port on which to look up the remote object.
    */
   public TcpDelegatingCacheLoader(String host, int port) {
      this.host = host;
      this.port = port;
   }

   /**
    * Allows configuration via XML config file.
    * 
    * @see org.jboss.cache.loader.DelegatingCacheLoader#setConfig(java.util.Properties)
    */
   public void setConfig(Properties props) {
      this.host = props.getProperty("host");
      if(this.host == null || this.host.length() == 0) {
         this.host = "localhost";
      }
      this.port = Integer.parseInt(props.getProperty("port"));
   }

   public void start() throws Exception {
      init();
   }

   public void stop() {
      try {if(in != null) in.close();} catch(IOException e) {}
      try {if(out != null) out.close();} catch(IOException e) {}
      try {if(sock != null) sock.close();} catch(IOException e) {}
   }


   private void init() throws IOException {
      if(host == null)
         host="localhost";
      sock=new Socket(host, port);
      out=new ObjectOutputStream(sock.getOutputStream());
      in=new ObjectInputStream(sock.getInputStream());
   }

   /**
    * Allows configuration via XML config file.
    * 
    * @see org.jboss.cache.loader.DelegatingCacheLoader#setCache(org.jboss.cache.TreeCache)
    */
   public void setCache(TreeCache cache) {
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateGetChildrenNames(org.jboss.cache.Fqn)
    */
   protected Set delegateGetChildrenNames(Fqn fqn) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateGetChildrenNames);
      out.writeObject(fqn);
      return (Set)in.readObject();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateGet(org.jboss.cache.Fqn, Object)
    */
   protected Object delegateGet(Fqn name, Object key) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateGetKey);
      out.writeObject(name);
      out.writeObject(key);
      return in.readObject();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateGet(org.jboss.cache.Fqn)
    */
   protected Map delegateGet(Fqn name) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateGet);
      out.writeObject(name);
      return (Map)in.readObject();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateExists(org.jboss.cache.Fqn)
    */
   protected boolean delegateExists(Fqn name) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateExists);
      out.writeObject(name);
      return in.readBoolean();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegatePut(org.jboss.cache.Fqn, Object, Object)
    */
   protected Object delegatePut(Fqn name, Object key, Object value) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegatePutKeyVal);
      out.writeObject(name);
      out.writeObject(key);
      out.writeObject(value);
      return in.readObject();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegatePut(org.jboss.cache.Fqn, java.util.Map)
    */
   protected void delegatePut(Fqn name, Map attributes) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegatePut);
      out.writeObject(name);
      out.writeObject(attributes);
      out.flush();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateRemove(org.jboss.cache.Fqn, Object)
    */
   protected Object delegateRemove(Fqn name, Object key) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateRemoveKey);
      out.writeObject(name);
      out.writeObject(key);
      return in.readObject();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateRemove(org.jboss.cache.Fqn)
    */
   protected void delegateRemove(Fqn name) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateRemove);
      out.writeObject(name);
      out.flush();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateRemoveData(org.jboss.cache.Fqn)
    */
   protected void delegateRemoveData(Fqn name) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateRemoveData);
      out.writeObject(name);
      out.flush();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateLoadEntireState()
    */
   public byte[] delegateLoadEntireState() throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateLoadEntireState);
      return (byte[])in.readObject();
   }

   /**
    * @see org.jboss.cache.loader.DelegatingCacheLoader#delegateStoreEntireState(byte[])
    */
   public void delegateStoreEntireState(byte[] state) throws Exception {
      out.writeInt(DelegatingCacheLoader.delegateStoreEntireState);
      out.writeObject(state);
      out.flush();
   }



}
